"""
Kurtosis-specific routines
"""
import warnings

import numpy as np
from scipy.signal import lfilter
from obspy.core.stream import Stream
from obspy.core.trace import Trace
# from scipy.stats import kurtosis as scipy_kurtosis
# from obspy.realtime.signal import kurtosis as obspy_kurtosis

from .utils import smooth_filter
from .pick_candidate import PickCandidate
from .logger import log


class Kurtosis_pick():
    """
    Holds one Kurtosis-based pick and the values that chose it
    """
    def __init__(self, time, k_jump, kurtosis):
        """
        :param time: kurtosis pick time
        :param k_jump: jump in kurtosis associated with the pick
        :param kurtosis: kurtosis value at pick
        """
        self.time = time
        self.k_jump = k_jump
        self.kurtosis = kurtosis


class Kurtosis():
    """
    Class for picking seismograms using the Kurtosis

    Kurtosis picking is a n-step process whose goal is to identify jumps
    in the Kurtosis:
    1) calculate the Kurtosis
    2) calculate the cumulative sum of the kurtosis (kurto_cum)
    3) remove a linear fit from the beginning to the end of the kurto_cum
       (corr_kurto_cum)
    4) Identify peaks (highs) and subtract this from the preceding data.  Set
       all positive values to zero
    5) identify where the slope of the resulting function changes from positive
       to negative
    """
    def __init__(self, params, plot=False):
        """
        :param params: KurtosisParameters object
        :param plot: make a plot of Kurtosis parameters
        """
        self.params = params
        # self.freq_bands = k_params.frequency_bands
        # self.wind_lengths = k_params.window_lengths
        # self.extrem_smooths = k_params.extrema_smoothings
        # self.n_smooth = n_smooth
        self.plot = plot

        # Mean trace over freq_bands, wind_lengths & smoothing
        self.mean_kurtosis = None
        # Mean trace, cumulated and detrended
        self.mean_cumulative_kurtosis = None
        # Gradients of mean_cumulative_kurtosis, for each extrem_smoothing
        self.kurto_gradients = None

    def __str__(self):
        s = "Kurtosis:\n"
        s += f"   params = {self.params}\n"
        s += f"   mean_kurtosis = {self.mean_kurtosis}\n"
        s += f"   mean_cumulative kurtosis = {self.mean_cumulative_kurtosis}\n"
        s += f"   len{self.kurto_gradients} kurto_gradients"
        return s

    def pick_trace(self, trace, n_candidates, starttime=None, endtime=None,
                   extrem_type='mini', extrem_normalize=False,
                   extrem_which='max'):
        """
        Pick a trace using the Kurtosis

        :param trace: waveform trace
        :param n_candidates: maximum number of candidates to return
        :param starttime: first time of interest
        :param endtime: last time of interest
        :param extrem_type: 'mini' or 'maxi', depending on wheter you want to
            follow minima or maxima
        :param extrem_normalize: normalize the cumulative Kurtosis gradient
        :param extrem_which:
            'first': select first 'max_cadidates' extrema > 0.1, order right to
                     left
            'max': select the 'max_candidates' biggest extrema, order biggest
                       to smallest
        :returns:  list of PickCandidates
        """
        self.calc_kurtocum(trace, starttime, endtime)
        candidates = self.follow_extrem(ext_type=extrem_type,
                                        max_candidates=n_candidates,
                                        normalize=extrem_normalize,
                                        order=extrem_which)
        return candidates

    def calc_kurtocum(self, trace, starttime=None, endtime=None, debug=False):
        """
        Calculate cumulative kurtosis over bandwidths and windows

        Puts mean kurtosis over all windows and freq_bands in
        self.mean_cumulative_kurtosis
        :param trace: the raw trace (works one trace, not a stream)
        :param starttime: is the first time of interest
        :param endtime: is the last time of interest
        """
        # trace2FWkurto.m:26
        assert isinstance(trace, Trace), "trace is not an obspy Trace"

        # sr = trace.stats.sampling_rate
        # data_length = trace.stats.endtime - trace.stats.starttime
        if starttime is None:
            starttime = trace.stats.starttime
        elif starttime < trace.stats.starttime:
            starttime < trace.stats.starttime
        if endtime is None:
            endtime = trace.stats.endtime
        elif endtime > trace.stats.endtime:
            endtime = trace.stats.endtime

        # Filter traces in different frequency bands
        B = []
        log('Pre-filtering data for kurtosis in {} bands'.format(
            self.params.frequency_bands), 'debug')
        for FB in self.params.frequency_bands:
            f = trace.copy()
            f.detrend('demean')
            # RAMP UP SIGNAL BEFORE FILTERING
            # ramp_time = 2 / FB[0]
            # ramp_pts = int(max(ramp_time * f.stats.sampling_rate, 50))
            # ramp = np.arange(0, ramp_pts) / ramp_pts
            # f.data[:ramp_pts] *= ramp
            f.filter(type='bandpass', freqmin=FB[0], freqmax=FB[1], corners=3)
            f = f.slice(starttime, endtime)
            # RAMP UP SIGNAL AFTER FILTERING
            # f.data[:ramp_pts] *= ramp
            B.append(f)

        # 2-level lists: : 1st dim: window_lengths, 2nd dim: freq bands
        K, C = [], []  # all kurtoses and all cumulative, detrended kurtoses
        log('Calculating kurtosis of filtered data using {}-s windows'
            .format(self.params.window_lengths), 'debug')
        for win_len in self.params.window_lengths:
            corr_cums, kurtos = self.calc_cum_kurtoses(B, win_len)
            C.extend(corr_cums)
            K.extend(kurtos)
        self.mean_kurtosis = _mean_trace(K)
        self.mean_cumulative_kurtosis = _mean_trace(C)

    def calc_cum_kurtoses(self, B, win_len):
        """
        Calculate kurtoses and cumulative kurtoses for given window length

        :param B: list of traces, filtered in different bands
        :win_len: window length in seconds
        """
        corr_cums, kurtos = [], []
        for tr, fb in zip(B, self.params.frequency_bands):
            sr = tr.stats.sampling_rate
            dl = tr.stats.endtime - tr.stats.starttime
            if win_len > dl:
                warnings.warn('Kurtosis window > data window ('
                              f'{win_len:.3g}s > {dl:.3g}s), skipping!')
                continue
            win_samps = min(int(np.floor(win_len * sr)) + 1, len(tr.data))
            k = _fast_kurtosis(tr, win_samps)
            filt = smooth_filter(k, self.params.n_smooth)
            corr_cum, _ = _f_cumul(filt.copy())
            corr_cum.detrend('simple')
            if self.plot:
                self._plot_kurtosis(win_len, fb, tr, filt, corr_cum)
            kurtos.append(k)
            corr_cums.append(corr_cum)
        return corr_cums, kurtos

    @staticmethod
    def _plot_kurtosis(wl, fb, trace, kurtosis, corr_cum):
        log(f'Plot kurtosis for win_len={wl}, f_band={fb}', 'debug')
        cor = corr_cum.copy()
        cor.stats.channel = 'COR'
        kurtosis.stats.channel = 'KUR'
        Stream([trace, kurtosis, cor]).plot(
            size=(600, 600), equal_scale=False)

    def follow_extrem(self, ext_type='mini', max_candidates=2,
                      normalize=False, order=None, debug=False):
        """
        Return extrema of the cumulative kurtosis

        Steps:
          1) Smooth the cumulative kurtosis (self.mean_cumulative_kurtosis)
             using the windows specified in self.params.extrema_smoothings
          2) locate the first 'n' first extremas of the smoothest function
          3) refine these extrema step by step through less and less smooth
             functions

        Uses self.mean_cumulative kurtosis
        :param ext_type: 'mini' or 'maxi':  follow minima or maxima
        :param max_candidates: max number of candidates to return
        :param normalize: normalize the gradient?
        :param order:
            'first': select first 'max_candidates' extrema > 0.1, ordered right
                     to left
            'max': select from biggest to smallest
        :returns:  list of PickCandidates
        """
        # follow_extrem.m:29
        # Parameters
        assert not len(self.mean_cumulative_kurtosis) == 0,\
            'no mean cumulative kurtosis!'
        st = self.mean_cumulative_kurtosis.stats.starttime
        sr = self.mean_cumulative_kurtosis.stats.sampling_rate

        all_extrema = self._get_extrema(ext_type, normalize)

        selected_extrema = _select_extrema(all_extrema[0],
                                           max_candidates,
                                           order)

        # Sharpen the indices/values using the smaller smoothing values
        sharp_extrema = []
        for e in selected_extrema:
            extrema = e
            for finer_extrema in all_extrema[1:]:
                c_extrem = _find_close_extrema(
                    extrema, finer_extrema, 40)
                if c_extrem is not None:
                    extrema = c_extrem
            sharp_extrema.append(extrema)

        return [PickCandidate(st + x['index']/sr,
                              'kurtosis',
                              x['value'],
                              sampling_rate=sr)
                for x in sharp_extrema]

    def _get_extrema(self, ext_type, normalize, debug=False):
        """
        Reverse sorts extrema_smoothings and returns corresponding extrema

        :param ext_type: 'maxi' or 'mini', passed on to ext_indices
        :param normalize: passed on to _cum2grad()
        :returns: list of lists of {indices: value, trace: value} from
            smoothest to roughest.  value is approx height of the kurtosis jump
        """
        extrema, gradients = [], []
        # Put the strongest smoothing first
        self.params.extrema_smoothings.sort(reverse=True)
        # sorted_smooth = sorted(self.params.extrema_smoothings, reverse=True)
        # self.params.extrema_smoothings = sorted_smooth = sorted(
        for smoothing in self.params.extrema_smoothings:
            v_smooth = smooth_filter(self.mean_cumulative_kurtosis, smoothing)
            v_smooth = _cum2grad(v_smooth, normalize)
            ext_indices = _loca_ext(v_smooth, ext_type)
            extrema.append([{'index': i, 'value': -v_smooth.data[i]}
                            for i in ext_indices])
            gradients.append(v_smooth)
        self.kurto_gradients = gradients
        return extrema


def _find_close_extrema(best_extrem, finer_extrema, max_diff=40):
    if len(finer_extrema) == 0:
        warnings.warn('No extrema found for {:d}-sample smoothing'.format(
                      finer_extrema['smoothing']))
        return None
    i_diff = abs(best_extrem['index'] - [x['index'] for x in finer_extrema])
    if np.any(i_diff <= max_diff):
        return finer_extrema[np.argmin(i_diff)]
    return None


def _select_extrema(extrema, N, order='max', threshold=0.1):
    """
    Return N extrema, ordered by size or time
    :param extrema: list of {'value': val, 'index': i}
    :param order: 'max' return from largest to smallest
                  'first': return from first to last
    :param threshold: minimum value to accept for ex=='first'
    """
    if order == 'first':
        big_extrema = [x for x in extrema if x['value'] >= threshold]
        if len(big_extrema) > 1:
            ext = sorted(big_extrema, key=lambda k: k['index'])
        else:
            ext = sorted(extrema, key=lambda k: k['index'])
        try:
            selected = [x for x in ext[:N:-1]]
        except Exception:
            selected = [x for x in ext[::-1]]
    elif order == 'max':
        ext = sorted(extrema, key=lambda k: np.abs(k['value']), reverse=True)
        try:
            selected = [x for x in ext[:N]]
        except Exception:
            selected = [x for x in ext]
    else:
        raise NameError("order not 'max' or 'first'")
    return selected


def _fast_kurtosis(trace, win_samps):
    """
    Compute kurtosis quickly using "filter" function

    could I just use obspy kurtosis?

    :param trace: one trace
    :param win_samps: number of samples in the sliding window
    :returns: Kurtosis trace
    """
    assert isinstance(trace, Trace), "trace is not an obspy Trace"
    out = trace.copy()
    win_samps = int(round(win_samps))
    # log(f'{win_samps=}', 'debug')
    if win_samps == 1:
        win_samps = 2
    a = np.divide(np.ones(win_samps), float(win_samps))
    b = 1.

    f = trace.copy()
    f.detrend(type='demean')

    # Compute kurtosis
    m_2 = lfilter(a, b, f.data**2)
    m_4 = lfilter(a, b, f.data**4)
    out.data = np.divide(m_4, (m_2 ** 2))
    # Protect against edge effect
    out.data[:win_samps] = out.data[win_samps]

    # Set any kurtosis value to nan for any indices within win_samples of
    # an NaN in the original data.
    # I think this is outdated, should just trim
    # for i in np.nonzero(trace.data == np.nan)[0]:
    #     if i:
    #         out.data[i: i + win_samps] = np.nan
    return out


def _f_cumul(f):
    """
    Calculate the positive gradient cumulative of f

    If the gradient is positive we take the cumulative, if the gradient is
    negative then, as long as the gradient is negative, the value is that
    of the last positive gradient. The output has then only positive
    gradients
         ___/
        /
    ___/
    :param f: trace, stream or list of traces containing the data
    :returns: trace or list of cumulative output traces, first value = 0
              list of cumulative output traces
    """
    # f_cumul.m:14
    bare_trace = False
    if isinstance(f, Trace):
        f = [f]
        bare_trace = True
    g = f.copy()
    for t in g:
        # tdata = t.data.copy()    # Backup copy for info
        t.data[np.isnan(t.data)] = 0
        t.data = np.gradient(t.data)
        # t.data = np.diff(t.data, prepend=0)
        # t.differentiate(method='gradient')
        t.data[t.data < 0] = 0
        t.data = np.cumsum(t.data)
        # t.data[np.isnan(tdata)] = np.nan

    p = g.copy()

    # Subtract first non-nan value from g
    for t in g:
        ind = np.nonzero(np.isfinite(t.data))[0]
        if len(ind) > 0:
            t.data -= t.data[ind[0]]

    if bare_trace:
        return g[0], p[0]
    else:
        return g, p


def _f_segment(f):
    """
    Return a line segment between the first and last values of function.
    SHOULD BE REPLACEABLE BY DETREND('linear')

    Goes from first non-nan value to last non-nan value
    Input and output have the same size.
    :param f: trace, stream or list of data traces
    :returns: line segment trace or list of traces
    """
    bare_trace = False
    if isinstance(f, Trace):
        f = [f]
        bare_trace = True
    assert isinstance(f[0], Trace), f'f is list of {type(f[0])}s!'
    # f_segment.m:12
    segments = []
    # for i in arange(1,n).reshape(-1):
    for trace in f:
        a = np.nonzero(np.isfinite(trace.data))[0][0]
        b = np.nonzero(np.isfinite(trace.data))[0][-1]
        ya = trace.data[a]
        yb = trace.data[b]
        lin = np.linspace(start=ya, stop=yb, num=b-a+1)
        segment = trace.copy()
        # Ugly handling of different cases
        before = np.empty(a)
        before[:] = np.nan
        after = np.empty(len(trace.data)-b-1)
        after[:] = np.nan
        segment.data = np.concatenate((before, lin, after))
        segments.append(segment)
    if bare_trace:
        return segments[0]
    else:
        return segments


def _cum2grad(f_in, normalize=False, debug=False):
    """
    Transform cumulative into function that is shifted below 0 for all maxima

    Finds peaks, subtracts the next peak value from all data, gets rid of
    all values > 0.  Give negative values that "jump" up to zero at the next
    peak

    :param f_in: cumulative trace
    :param normalize: 'True' will divide the output by it's min value
    :returns: f_out
    """
    # cum2grad.m:7
    assert not len(f_in) == 0, 'f_in is empty!'

    tycalpha = np.zeros(len(f_in.data))
    tikxs = _loca_ext(f_in, 'maxi')
    # a = int(np.nonzero(np.isfinite(f_in.data))[0][0])
    # b = int(np.nonzero(np.isfinite(f_in.data))[0][-1])
    a = 0
    b = len(f_in.data) - 1
    # print(f'{f_in.data.shape=}, {tycalpha.shape=}, {b=}')
    # print(f'{tikxs=}')
    # tikxs = tikxs.to_list()
    if len(tikxs) == 0:
        tikxs = np.array([a, b])
    else:
        tikxs = np.array([a] + tikxs.tolist() + [b])

    # function equal to the next peak
    for j in range(len(tikxs)-2, -1, -1):
        tycalpha[tikxs[j]:tikxs[j+1]+1] = f_in.data[tikxs[j+1]]

    f_out = f_in.copy()
    f_out.data -= tycalpha          # input minus the next peak value
    f_out.data[f_out.data > 0] = 0  # Get rid of everything above zero
    if normalize:
        f_out.data /= abs(min(f_out.data))
    if debug:
        Stream([f_in, f_out]).plot()
    return f_out


def _loca_ext(trace, ext_type, starttime=None, endtime=None, debug=False):
    """
    Returns local extrema

    Actually just returns where the trace slope changes from positive to
    negative (ext_type=='maxi'), or vice versa (type='mini')
    :param trace: waveform trace
    :param ext_type: 'maxi' or 'mini'
    :param start_time: start of window to look at
    :param end_time: end of window to look at
    :returns: indices
    """
    assert ext_type in ('mini', 'maxi')
    diff = trace.copy()
    diff.data = np.diff(np.sign(np.diff(trace.data)), prepend=0)
    if starttime is not None or endtime is not None:
        diff.trim(starttime, endtime)
    if debug:
        diff.plot()
    if ext_type == 'maxi':
        loc = (diff.data < 0)
    else:
        loc = (diff.data > 0)

    i_extremes = np.nonzero(loc)[0] + 1
    return i_extremes


def _mean_trace(traces):
    """
    Calculate the mean trace

    :param traces: stream or list of traces
    :returns: trace object
    """
    if isinstance(traces, Trace):
        return traces.copy()
    elif isinstance(traces[0], Trace):
        if len(traces) == 1:
            return traces[0].copy()
    else:
        ValueError("traces is not a Trace or a list of Traces")

    s = Stream(traces).copy()
    s = s.stack(npts_tol=1, time_tol=1./traces[0].stats.sampling_rate)
    if s[0].stats.starttime.timestamp == 0:
        ValueError('input traces did not start at same time')
    return s[0]
    # data_len = len(traces[0].data)
    # for tr in traces[1:]:
    #     assert len(tr.data) == data_len, 'traces are not the same length'
    #
    # log(f'Calculating mean of {len(traces):d} traces', 'debug')
    # mean_tr = traces[0].copy()
    # for tr in traces[1:]:
    #     mean_tr.data += tr.data
    # mean_tr.data /= len(traces)
    # return mean_tr


if __name__ == '__main__':
    pass



# Generated with SMOP  0.41-beta
# from smop.libsmop import *
# readmain.m
import yaml

from .channel_mapping_rules import ChannelMappingRules
from .station_parameters import StationParameters
from .SNR_parameters import SNRParameters
from .polarity_parameters import PolarityParameters
from .associator_parameters import AssociatorParameters
from .global_window_parameters import GlobalWindowParameters
from ..timer import Timer


class PickerParameters():
    """
    Picker Parameters
    """
    def __init__(self, global_window, SNR, polarity={}, channel_parameters={},
                 association={}, response_file_type='', station_parameters={},
                 stations={}):
        """
        Initialize Picker Parameters

        :param gw: GlobalWindowParameters dictionary source
        :param SNR: SNRParameters dictionary source
        :param polarity: PolarityParameters dict source
        :param assoc: AssociatorParameters dict source
        :param station_parameters: dict with key=station type,
                                   items=StationParameters object
        :param stations: dict with key=station name,
                                   items={station type, resp_file_name}
        :param channel_parameters: ChannelMappingRules dictionary source
        :param response_file_type: Format of the response file(s).  ['GSE',
            'JSON_PZ', 'SACPZ', 'STATIONXML' or
            empty.  If empty use Baillardesque PoleZero format.
        """
        self.gw = GlobalWindowParameters(**global_window)
        self.SNR = SNRParameters(**SNR)
        self.polarity = PolarityParameters(**polarity)
        self.channel_mapping_rules = ChannelMappingRules(**channel_parameters)
        self.assoc = AssociatorParameters(**association)
        self.response_file_type = response_file_type

        self.station_parameters = {}
        for station, values in stations.items():
            temp = station_parameters[values['parameters']]
            temp['resp_file'] = values['resp_file']
            self.station_parameters[station] = StationParameters.from_dict(
                temp)

    @property
    def stations(self):
        return [s for s in self.station_parameters.keys()]

    @property
    def n_stations(self):
        return len(self.station_parameters)

    def __str__(self):
        str = "PickerParameters:\n"
        str += f"    gw = {self.gw}\n"
        str += f"    SNR = {self.SNR}\n"
        str += f"    polarity = {self.polarity}\n"
        str += f"    assoc = {self.assoc}\n"
        str += f"    stations = {','.join(self.stations)}\n"
        str += f"    response_file_type = '{self.response_file_type}'\n"
        str += f"    channel_mapping_rules = {self.channel_mapping_rules}\n"
        return str

    @classmethod
    def from_yaml_file(cls, filename):
        """
        Read in parameters from a YAML file
        """
        with Timer(text="Read picker parameter file: {:0.4f}s"):
            with open(filename, 'r') as fic:
                params = yaml.safe_load(fic)

            return cls(**params)


if __name__ == '__main__':
    pass



import logging
import verboselogs
# import sys
import inspect
from datetime import datetime

# from obspy.core import UTCDateTime
#
logger_name = 'pspicker'


# Set up Filter to not output logs from other programs to file handler
# (because it requires the "caller" variable)
class NoParsingFilter(logging.Filter):
    def filter(self, record):
        return record.name == logger_name


def setup_log(stream_log_level=logging.INFO,
              file_log_level=logging.VERBOSE):
    """
    Set console and log file logging levels

    :param stream_log_level: log level for console output
    :kind stream_log_level: verboselogs level, str or None
    :param file_log_level: log level for log file.  If higher than the
        stream_log_level, will be set to stream_log_level

    For information, the log levels are:
        DEBUG: 10
        VERBOSE: 15
        INFO: 20
        WARNING: 30
        ERROR: 40
        CRITICAL: 50
    """
    if stream_log_level is None:
        stream_log_level = logging.INFO
    elif isinstance(stream_log_level, str):
        stream_log_level = getattr(logging, stream_log_level.upper())
    if stream_log_level not in [logging.DEBUG, logging.VERBOSE, logging.INFO,
                                logging.WARNING, logging.ERROR,
                                logging.CRITICAL]:
        stream_log_level = logging.INFO
    ts = datetime.today().strftime('%Y%m%dT%H%M')

    file_log_level = min(file_log_level, stream_log_level)

    verboselogs.install()   # Sets VerboseLogger as the default Logger

    # Set up Formatters
    lf = logging.Formatter('%(asctime)s %(levelname)-8s - '
                           '%(message)s (in %(caller)s)')
    cf = logging.Formatter('%(levelname)-8s %(message)s')

    # Set up Handlers
    lh = logging.FileHandler(f'run_{ts}.log', 'w')
    lh.setLevel(file_log_level)
    lh.setFormatter(lf)
    lh.addFilter(NoParsingFilter())
    ch = logging.StreamHandler()
    ch.setLevel(stream_log_level)
    ch.setFormatter(cf)

    global logger
    logger = logging.getLogger(logger_name)
    logger.setLevel(file_log_level)  # THIS IS NECESSARY!!!!
    logger.addHandler(lh)
    logger.addHandler(ch)
    logging.captureWarnings(True)
    print(f'{logger=}')
    print(f'{logger.handlers=}')


def log(string, level="info"):
    """
    Prints a string and logs to file

    :param level: the log level
    """
    caller = inspect.stack()[1]
    caller_str = caller[3] + '()'
    level = level.upper()
    extra = {'caller': caller_str}

    level = getattr(logging, level.upper())

    global logger
    logger.log(level, string, extra=extra)



# from readMSEEDTraces.m
class ChannelMap():
    """
    Class to store seed_ids corresponing to diferent components
    """
    def __init__(self, Z, N, E, H, P_write_cmp, P_write_phase,
                 S_write_cmp, S_write_phase):
        """
        :paran Z: seed_id corresponding to station's vertical component
        :paran N: seed_id corresponding to station's north component
        :paran E: seed_id corresponding to station's east component
        :paran H: seed_id corresponding to station's hydrophone/pressure
            component
        :param P_write_cmp: component to write P picks to
        :param S_write_cmp: component to write S picks to
        :param P_write_phase: what phase to name P picks
        :param S_write_phase: what phase to name S picks
        """
        assert P_write_cmp in "ZNEH", 'not in "ZNEH"'
        assert S_write_cmp in "ZNEH", 'not in "ZNEH"'
        assert Z is not None
        self.Z = Z
        self.N = N
        self.E = E
        self.H = H
        self._P_write_cmp = P_write_cmp
        self._S_write_cmp = S_write_cmp
        self.P_write_phase = P_write_phase
        self.S_write_phase = S_write_phase

    @property
    def P_write_seed_string(self):
        """ Returns the full seed string name"""
        return getattr(self, self._P_write_cmp)

    @property
    def S_write_seed_string(self):
        """ Returns the full seed string """
        return getattr(self, self._S_write_cmp)

    @property
    def P_write_cmp(self):
        """ Returns the component name"""
        return self.P_write_seed_string.split('.')[-1]

    @property
    def S_write_cmp(self):
        """ Returns the component name"""
        return self.S_write_seed_string.split('.')[-1]

    def __str__(self, format=None):
        """
        :param format: 'table_row': print as a table row
                       'table_header': print a table header (no object info)
                       None: print normally
        """
        if format == 'table_header':
            s = '{:^6s}|{:^6s}|{:^6s}|{:^6s}|{:^9s}|{:^9s}|{:^8s}|{:^8s}|'.\
                format('Z', 'N', 'E', 'H', 'P_write', 'S_write',
                       'P_phase', 'S_phase')
            # s += '{:-<6s}+{:-<6s}+{:-<6s}+{:-<6s}+'.format(
            #     '', '', '', '')
            # s += '{:-<9s}+{:-<9s}+{:-<8s}+{:-<8s}|'.format(
            #     '', '', '', '')

            return s
        elif format == 'table_row':
            H, Z, N, E = 'None', 'None', 'None', 'None'
            if self.H is not None:
                H = self.H.split('.')[-1]
            if self.Z is not None:
                Z = self.Z.split('.')[-1]
            if self.N is not None:
                N = self.N.split('.')[-1]
            if self.E is not None:
                E = self.E.split('.')[-1]
            s = '{:^6s}|{:^6s}|{:^6s}|{:^6s}|{:^9s}|{:^9s}|{:^8s}|{:^8s}|'.\
                format(Z, N, E, H,
                       self.P_write_cmp, self.S_write_cmp,
                       self.P_write_phase, self.S_write_phase)
#                       self._write_cmp_str(self._P_write_cmp),
#                       self._write_cmp_str(self._S_write_cmp),
#                       self.P_write_phase, self.S_write_phase)
            return s
        else:
            s = f'            Z: {str(self.Z)}\n'
            s += f'            N: {str(self.N)}\n'
            s += f'            E: {str(self.E)}\n'
            s += f'            H: {str(self.H)}\n'
            s += f'  P_write_cmp: {self._P_write_cmp} (=>{self.P_write_cmp})\n'
            s += f'  S_write_cmp: {self._S_write_cmp} (=>{self.S_write_cmp})\n'
            s += f'P_write_phase: {self.P_write_phase}\n'
            s += f'S_write_phase: {self.S_write_phase}\n'
            return s

    def _shortChName(self, write_to):
        """Shortened channel name of the given seed_id"""
        seed_id = self[write_to]     # Dict acces to class parameter
        return seed_id[-3] + seed_id[-1]



"""
Pick candidate class
"""
from obspy.core import UTCDateTime
from obspy.core.event.origin import Pick, Arrival
from obspy.core.event.base import WaveformStreamID, QuantityError

from .logger import log


class PickCandidate():
    def __init__(self, time, picker_type, picker_value, snr=None,
                 DR=None, phase_guess=None, sampling_rate=None, station=None,
                 weight=None):
        """
        :param time: time of the PickCandidate
        :param picker_type: picker type (e.g. 'kurtosis')
        :param picker_value: measure of pick quality (higer=better) made
            by the picker
        :param snr: signal-to-noise ratio at the pick
        :param DR: Dip-Rect parameter associated with the PickCandidate
        :param phase_guess: most likely phase type
        :param station: station for this candidate
        :param sampling_rate: sampling rate of data used to make the candidate
        :param weight: Nordic-style pick weight [0=best, 3=worst]
        """
        assert isinstance(time, UTCDateTime), 'time is not a UTCDateTime'
        assert isinstance(picker_type, str), 'picker_type is not a str'
        picker_value = float(picker_value)

        self.time = time
        self.picker_type = picker_type
        self.picker_value = picker_value
        self.snr = snr
        self.DR = DR
        self.phase_guess = phase_guess
        self.sampling_rate = sampling_rate
        self.station = station
        self.weight = weight

    def __str__(self):
        s = f"time={self.time.strftime('%Y%m%dT%H%M%S')}"
        s += f", {self.picker_type}, value={self.picker_value:.3g}"
        if self.snr is not None:
            s += f", snr={self.snr:.3g}"
        if self.DR is not None:
            s += f", DR={self.DR:.3g}"
        if self.phase_guess is not None:
            s += f', phase_guess="{self.phase_guess}"'
        if self.sampling_rate is not None:
            s += f', sampling_rate={self.sampling_rate:.3g}'
        if self.station is not None:
            s += f', station="{self.station}"'
        return s

    def __eq__(self, other):
        if not self.time == other.time:
            return False
        if not self.picker_type == other.picker_type:
            return False
        if not self.picker_value == other.picker_value:
            return False
        if not self.snr == other.snr:
            return False
        if not self.DR == other.DR:
            return False
        if not self.phase_guess == other.phase_guess:
            return False
        if not self.sampling_rate == other.sampling_rate:
            return False
        if not self.station == other.station:
            return False
        if not self.weight == other.weight:
            return False
        return True

    @property
    def shortname(self):
        return f'{self.station}_{self.phase_guess}'

    def to_obspy(self, channel_maps, quality_thresholds=None):
        """
        Return obspy Pick and Arrival

        Arrival is None if there is no pick weight
        :param channel_maps: {station: ChannelMap object}
        :param quality_thresholds: list of 4 signal-to-noise ratio thresholds
            for pick quality, from lowest [quality=3] to highest [quality=0]
        :parm sampling_rate: sampling rate (used to estimate timing error)
        """
        c_map = channel_maps[self.station]
        if self.phase_guess == 'P':
            id = WaveformStreamID(seed_string=c_map.P_write_seed_string)
            phase_hint = c_map.P_write_phase
            log(f'{phase_hint}: {id.get_seed_string()}', 'verbose')
        elif self.phase_guess == 'S':
            id = WaveformStreamID(seed_string=c_map.S_write_seed_string)
            phase_hint = c_map.S_write_phase
            log(f'{phase_hint}: {id.get_seed_string()}', 'verbose')
        else:
            raise ValueError("phase guess '{self.phase_guess}' not 'P' or 'S'")
        self.weight = self._get_weight(quality_thresholds)
        time_errors = self._time_error()
        pick = Pick(time=self.time,
                    time_errors=time_errors,
                    waveform_id=id,
                    phase_hint=phase_hint,
                    evaluation_mode='automatic',
                    evaluation_status='preliminary')
        if self.weight is not None:
            return pick, Arrival(pick_id=pick.resource_id,
                                 time_weight=self.weight)
        else:
            return pick, None

    def _get_weight(self, quality_thresholds):
        """
        Return pick weights corresponding to SNR

        :param quality_thresholds: list of 4 signal-to-noise ratio thresholds
            for pick quality, from lowest [quality=3] to highest [quality=0]
        :returns: pick quality (0 to 3)
        :rtype: obspy QuantityError
        """
        if quality_thresholds is None:
            return None
        quality_thresholds.sort()
        if self.snr > quality_thresholds[3]:
            return 0
        elif self.snr >= quality_thresholds[2]:
            return 1
        elif self.snr >= quality_thresholds[1]:
            return 2
        elif self.snr >= quality_thresholds[0]:
            return 3
        else:
            return 4

    def _time_error(self,):
        """
        Return approximate pick time errors corresponding to pick weight

        Errors are multiplied by 2 for self.phase_guess == 'S'
        :returns: time_errors
        :rtype: obspy QuantityError
        """
        if not isinstance(self.weight, int):
            return None
        elif self.sampling_rate is None:
            return None

        assert self.phase_guess in 'PS',\
            "phase_guess '{self.phase_guess}' not in 'PS'"

        if self.weight == 0:
            uncertainty = 2. / self.sampling_rate
        elif self.weight == 1:
            uncertainty = 8. / self.sampling_rate
        elif self.weight == 2:
            uncertainty = 32. / self.sampling_rate
        elif self.weight == 3:
            uncertainty = 128. / self.sampling_rate
        else:
            uncertainty = 2000. / self.sampling_rate
        if self.phase_guess == 'S':
            uncertainty *= 2.
        return QuantityError(uncertainty)

    @staticmethod
    def remove_duplicates(pick_list):
        """
        Remove duplicate PickCandidates from a list

        Mostly a debugging routine
        """
        i_remove = []
        for i in range(len(pick_list)-1):
            for j in range(i+1, len(pick_list)):
                if pick_list[i] == pick_list[j]:
                    log(f'picks {i} & {j} are identical: {pick_list[i]}',
                        'warning')
                    i_remove.append(j)
        if len(i_remove) > 0:
            pick_list = [p for p, i in zip(pick_list, range(len(pick_list)))
                         if i not in i_remove]
        return pick_list



# Generated with SMOP  0.41-beta
# from smop.libsmop import *
import warnings

from obspy.core import Stream  # , Trace


class PickerStationParameters():
    """
    Parameters associated with work on a single station for one event
    """
    def __init__(self,
                 station=None,
                 station_params=None,
                 channel_map=None,
                 stream=None):
        """
        Initiate loop parameters

        :param station: station name
        :param station_params: StationParameters object
        :param channel_map: ChannelMapping object
        :param stream: obspy Stream containing appropriate traces
        """
        self.station = station
        self.station_params = station_params
        self.channel_map = channel_map
        self.datP = self._get_traces(stream, station_params.P_comp)
        self.datS = self._get_traces(stream, station_params.S_comp)
        # print(self.datP)
        # print(self.datS)
        self.dat_noH = self._get_noH_traces(stream)
        self.data_limits = self._find_limits()
        self.t_begin = self._find_first_time()

    def _get_traces(self, stream, comp_list):
        """
        Adds datP or datS

        :param stream: Stream of all Traces
        :param comp_list: list of components to include in this stream
        """
        out_stream = Stream()
        # make sure 'Z' is at end of comp_list (for polarity analysis)
        if 'Z' in comp_list and len(comp_list) > 1:
            comp_list = comp_list.replace('Z', '') + 'Z'
        for comp in comp_list:
            st = stream.select(id=getattr(self.channel_map, comp))
            # st = stream.select(station=self.station, component=comp)
            if not len(st) == 1:
                txt = f'station={self.station}, comp={comp}'
                if len(st) == 0:
                    warnings.warn(f'No traces found for {txt}')
                    continue
                else:
                    warnings.warn('{:d)} traces found for {}, using first'.
                                  format(len(st), txt))
            out_stream += st[0]
        return out_stream

    def _get_noH_traces(self, stream):
        """
        Adds dat_noH
        """
        out_stream = Stream()
        for tr in stream.select(station=self.station):
            if not tr.stats.channel[-1] == 'H':
                out_stream += tr
        return out_stream

    def _find_limits(self):
        """
        find limits to self.datP, self.datS and self.dat_noH

        Uses only self.dat_noH because the others should be included
        """
        assert self.dat_noH is not None, 'self.dat_noH is None'
        assert len(self.dat_noH) > 0, 'no traces in self.dat_noH'
        dmin = 1.e99
        dmax = -1.e99
        for tr in self.dat_noH:
            if tr.data.max() > dmax:
                dmax = tr.data.max()
            if tr.data.min() < dmin:
                dmin = tr.data.min()
        return (dmin, dmax)

    def _find_first_time(self):
        """
        Finds first time of all traces

        Uses only self.dat_noH because the others should be included
        """
        assert self.dat_noH is not None, 'self.dat_noH is None'
        assert len(self.dat_noH) > 0, 'no traces in self.dat_noH'

        first_time = self.dat_noH[0].stats.starttime
        for tr in self.dat_noH[1:]:
            if tr.stats.starttime < first_time:
                first_time = tr.stats.starttime
        return first_time

    # def index_to_time(self, index, trace=None):
    #     """
    #     Returns corresponding to given index
    #
    #     index: the index
    #     trace: the trace from which to get the start time and sampling_rate.
    #         If None, use self.datP[0]
    #     """
    #     if trace is None:
    #         trace = self.datP[0]
    #     assert isinstance(trace, Trace), f'trace is {type(trace)}, not Trace'
    #     sr = trace.stats.sampling_rate
    #     t_begin = trace.stats.starttime
    #     return t_begin + index/sr


if __name__ == '__main__':
    pass



# import os.path
#
# import json

import numpy as np
from matplotlib import pyplot as plt
from matplotlib.patches import Rectangle
from scipy.signal import find_peaks
from obspy.core.event.magnitude import Amplitude
from obspy.core.event.origin import Pick
from obspy.signal.invsim import estimate_wood_anderson_amplitude

from .logger import log
from .paz import PAZ


class Amp():
    """Mini class to hold and print amplitude values"""
    def __init__(self, value=0, period=0, time=None, channel=None):
        self.value = value
        self.period = period
        self.time = time
        self.channel = channel

    def __str__(self):
        return "Amp: value={:.3g}, period={:.3g}, time={}, chan={}".format(
            self.value, self.period, self.time, self.channel)


class LocalAmplitude():
    """
    Calculate local earthquake signal amplitude at a station
    """
    def __init__(self, traces, picks, response_file, response_file_type,
                 P_pick_window=[-5, 30], S_pick_window=[-20, 10]):
        """
        :param traces: all traces associated with this station/event
        :param picks: P- and S- Picks
        :type picks: list of obspy Pick
        :param response_file: file containing instrument response
        :param response_file_type:
        :param P_pick_window: if there is only a P pick, window before and
            after to look for maximum
        :param P_pick_window: if there is only an s pick, window before and
            after to look for maximum
        """
        assert len(picks) > 0, "no picks!"
        for pick in picks:
            assert isinstance(pick, Pick), 'pick is not an obspy Pick'
        self.pick_P = self._get_pick(picks, "P")
        self.pick_S = self._get_pick(picks, "S")
        self.P_pick_window = P_pick_window
        self.S_pick_window = S_pick_window
        self.win_start, self.win_end, self.ref_pick, self.station =\
            self._set_ampl_window()
        # Should take advantage of obspy to read all standard formats
        self.paz = get_response(response_file, response_file_type)
        self.paz.input_units = 'nm'
        self.traces = traces

    def __str__(self):
        """
        """
        s = "LocalAmplitude: "
        if self.pick_P:
            s += f"P_time = {self.pick_P.time}, "
        if self.pick_S:
            s += f"S_time = {self.pick_S.time}, "
        if self.win_start:
            s += f"win_start = {self.win_start}, "
        if self.win_end:
            s += f"win_end = {self.win_end}, "
        if self.traces:
            s += "traces = ({}), ".format(
                ",".join([x.stats.channel for x in self.traces]))
        s += "paz = {}, ".format(self.paz)
        s += "P_pick_window = {}, ".format(self.P_pick_window)
        s += "S_pick_window = {}".format(self.S_pick_window)
        return s

    def get_iaml(self, plot=False, method='wood_calc'):
        """
        get IAML amplitude and associated pick
        From IAPEI CoSOI 2013 Working Group recommentations:
        "ML = log10(A) + 1.11log10(R) + 0.00189*R - 2.09
        where
        A = maximum *trace* amplitude in *nm* that is measured on output from
            a *horizontal-component* instrument that is filtered so that the
            response of the seismograph/filter system replicates that of a
            *Wood-Anderson standard seismograph* but with a
            static magnification of 1
        R = *hypocentral distance in km*, typically less than 1000 km

        the constant -2.09 is based on an experimentally determined static
        magnification of the Wood-Anderson of 2080, rather than the theoretical
        magnification of 2800.

        The amplitudes used in the magnitude formulas ... are in most
        circumstances to be measured as one-half the maximum deflection of the
        seismogram trace, peak-to-adjacent- trough or trough-to-adjacent-peak,
        where peak and trough are separated by one crossing of the zero-line:
        the measurement is sometimes described as “one-half peak-to-peak
        amplitude.” None of the magnitude formulas presented in this article
        are intended to be used with the full peak-to-trough deflection as
        the amplitude.

        WA displacement zeros = [(0+0j), (0+0j)]
        WA displacement poles = [(-5.49779 - 5.60886j), (-5.49779 + 5.60886j)]
        WA seismometer free period = 0.8 s
        WA damping constant = 0.7
        A0 = 0.97866 @ 4 Hz
        :param plot: plot the result
        :param method:
            "wood_calc": calculate amplitude on the signal transformed to
                         simulate a unity gain Wood-Anderson filter
            "wood_est": calculate amplitude on the original signal and use
                        obspy.signal.invsim.estimate_wood_anderson_amplitude()
            "raw_disp": calculate amplitude on the signal transformed to
                        displacement in nm
        :returns: amplitude in meters (type=IAML), associated Pick
        :rtype: ~class obspy.core.event.magnitude.Amplitude,
                ~class obspy.core.event.origin.Pick
        """
        if self.ref_pick is None:
            log('No ref_pick!', 'debug')
            return None
        assert method in ['wood_calc', 'wood_est', 'raw_disp']

        signal = self.traces.copy()
        # do all work in nm
        paz_simulate, paz_simulate_obspy = None, None
        paz_remove = self.paz.copy()
        if method == 'wood_est':
            paz_remove.input_units = 'm/s'
            plot_units = 'Original (counts)'
        else:
            if method == 'wood_calc':
                paz_simulate = PAZ(
                    poles=[(-5.49779 - 5.60886j), (-5.49779 + 5.60886j)],
                    zeros=[(0+0j), (0+0j)],
                    ref_gain=1, ref_freq=4.0,
                    input_units='nm', output_units='counts')
                plot_units = 'Wood-And (nm)'
                paz_simulate_obspy = paz_simulate.to_obspy()
            elif method == 'raw_disp':
                paz_simulate = None
                paz_remove.input_units = 'nm'
                plot_units = 'Disp (nm)'
            for tr in signal:
                tr.simulate(paz_remove=paz_remove.to_obspy(),
                            paz_simulate=paz_simulate_obspy,
                            water_level=60.0)
        amp = pk2pk(signal, self.win_start, self.win_end)
        if amp is None:
            return None, None
        if method == 'wood_est':
            amp.value = estimate_wood_anderson_amplitude(paz_remove.to_obspy(),
                                                         amp.value, amp.period)
            # estimate...() gives displacement amplitude ON a Wood-Anderson
            # seismometer, divide by 2080 then multiply by 1e6 to get ground
            # motion in nm?
            amp.value /= 2080./1e6
        if plot:
            self.plot(signal, amp, plot_units)
        waveform_id = self.ref_pick.waveform_id
        waveform_id.channel_code = amp.channel
        pick = Pick(time=amp.time,
                    waveform_id=waveform_id,
                    method_id='ps_picker',
                    phase_hint='IAML',
                    evaluation_mode='automatic',
                    evaluation_status='preliminary')
        obspy_amp = Amplitude(generic_amplitude=amp.value/(2. * 1.e9),
                              type='IAML',
                              unit='m',
                              period=amp.period,
                              magnitude_hint='ML',
                              category='period',
                              pick_id=pick.resource_id,
                              waveform_id=pick.waveform_id)
        return obspy_amp, pick

    def _set_ampl_window(self):
        if self.pick_S is not None and self.pick_P is not None:
            pick = self.pick_S
            win_start = self.pick_P.time - 1
            win_end = self.pick_S.time + 2*(self.pick_S.time
                                            - self.pick_P.time)
        elif self.pick_S is not None:
            pick = self.pick_S
            win_start = self.pick_S.time + self.S_pick_window[0]
            win_end = self.pick_S.time + self.S_pick_window[1]
        elif self.pick_P is not None:
            pick = self.pick_P
            win_start = self.pick_P.time + self.P_pick_window[0]
            win_end = self.pick_P.time + self.P_pick_window[1]
        else:
            return None, None, None, None
        return win_start, win_end, pick, pick.waveform_id.station_code

    @staticmethod
    def _get_pick(picks, phase):
        pick = [x for x in picks if x.phase_hint[0] == phase]
        if len(pick) == 0:
            return None
        else:
            if not len(pick) == 1:
                log('{:d} picks have phase[0] == "{}": {}, returning first one'
                    .format(len(pick), pick[0].phase_hint[0], pick), 'error')
            return pick[0]

    def plot(self, transformed, Amp, trans_units):
        """
        Plot the amplitude pick

        Top plot: all traces, colored by type
        Middle trace: Wood-Anderson traces, colored by type, with ampl pick
        Bottom trace: zoom on amplitude region, with max pick for max trace
        :param transformed: transformed traces
        :param Amp: pk2pk output
        :param trans_units: y-axis label for the transformed traces
        """
        fig, axs = plt.subplots(3, 1, num=f'LocalAmplitude {self.station}')
        traces = self.traces.slice(self.win_start, self.win_end)
        transcut = transformed.slice(self.win_start, self.win_end)
        imin = transcut[0].times('utcdatetime').searchsorted(Amp.time)
        imax = transcut[0].times('utcdatetime')\
                          .searchsorted(Amp.time + Amp.period) - 1
        zoom_start = Amp.time - 2*Amp.period
        zoom_end = Amp.time + 4*Amp.period

        # Top axis: plot all traces
        axs[0].set_xlim(self.win_start.matplotlib_date,
                        self.win_end.matplotlib_date)
        for tr in traces:
            color, label = self._choose_color(tr)
            axs[0].plot_date(tr.times(type="matplotlib"), tr.data, color + '-',
                             label=label)
        axs[0].add_patch(Rectangle((zoom_start.matplotlib_date,
                                    axs[0].viewLim.y0),
                                   width=(zoom_end - zoom_start)/86400,
                                   height=axs[0].viewLim.y1-axs[0].viewLim.y0,
                                   fc='blue', alpha=0.5, zorder=-1))
        if self.pick_P:
            axs[0].axvline(self.pick_P.time.matplotlib_date, color='b')
        if self.pick_S:
            axs[0].axvline(self.pick_S.time.matplotlib_date, color='r')
        axs[0].set_ylabel(self.station)
        axs[0].legend(loc='center left')

        # Middle axis: plot transformed traces
        axs[1].clear()
        axs[1].set_xlim(self.win_start.matplotlib_date,
                        self.win_end.matplotlib_date)
        for tr in transcut:
            color, label = self._choose_color(tr)
            axs[1].plot_date(tr.times(type="matplotlib"), tr.data, color + '-',
                             label=label)
        axs[1].add_patch(Rectangle((zoom_start.matplotlib_date,
                                    axs[1].viewLim.y0),
                                   width=(zoom_end - zoom_start)/86400,
                                   height=axs[1].viewLim.y1-axs[1].viewLim.y0,
                                   ec='blue', fc='lightblue', alpha=0.3,
                                   zorder=-1))
        axs[1].set_ylabel(trans_units)
        axs[1].legend(loc='center left')

        axs[2].set_xlim(zoom_start.matplotlib_date, zoom_end.matplotlib_date)
        tr = transcut.select(channel=Amp.channel)[0]
        color, label = self._choose_color(tr)
        axs[2].plot_date(tr.times(type="matplotlib"), tr.data, color + '-',
                         label=label)
        axs[1].plot(tr.times("matplotlib")[imin], tr.data[imin], color='k',
                    marker='x')
        axs[1].plot(tr.times("matplotlib")[imax], tr.data[imax], color='k',
                    marker='x')
        axs[2].plot(tr.times("matplotlib")[imin], tr.data[imin], color='k',
                    marker='x')
        axs[2].plot(tr.times("matplotlib")[imax], tr.data[imax], color='k',
                    marker='x')
        axs[2].set_ylabel(trans_units)

        plt.tight_layout()
        plt.draw()
        plt.show()
        # plt.show(block=False)
        plt.pause(0.001)

    @staticmethod
    def _choose_color(trace, Zcomps='Z3', Ncomps='N1Y', Ecomps='E2X'):
        """
        :param Zcomps: component characters to interpret as "Z" channel
        :param Ncomps: component characters to interpret as "N" channel
        :param Ecomps: component characters to interpret as "E" channel
        """
        comp = trace.stats.channel[-1]
        if comp in Ncomps:
            return 'r', 'N'
        elif comp in Ecomps:
            return 'b', 'E'
        elif comp in Zcomps:
            return 'k', 'Z'
        else:
            return 'g', comp


def pk2pk(stream, start_time, end_time):
    """
    Return the time, period and amplitude of the maximum peak-peak

    :param stream: waveform data
    :param pick_time: time of reference (usually a pick)
    :param start_time: start of amplitude window (UTCDateTime)
    :param end_time: end of amplitude window (UTCDateTime)
    :returns: Amp object
    """
    amp = Amp(value=0)
    if start_time > end_time:
        sta = list(set([tr.stats.station for tr in stream]))
        log(f'station {sta} pk2pk amplitude window start_time > end_time',
            'error')
        return None
    for tr in stream:
        window = tr.copy().trim(start_time, end_time, pad=True).data
        sr = tr.stats.sampling_rate
        # print(f'tr = {tr}')
        # print(f'window = {window}')
        i_mins, _ = find_peaks(-1. * window)
        i_maxs, _ = find_peaks(window)

        if len(i_mins) == 0 or len(i_maxs) == 0:
            return(None)

        # Make sure there is an i_max after the last i_min
        if i_mins[-1] == len(window)-1:
            i_mins = i_mins[:-1]
        if i_maxs[-1] <= i_mins[-1]:
            i_maxs = np.append(i_maxs, len(window)-1)

        for imin in i_mins:
            imax = np.min(i_maxs[i_maxs > imin])
            # print(imin, imax)
            amplitude = window[imax] - window[imin]
            if amplitude > amp.value:
                amp = Amp(amplitude, (imax - imin) / sr,
                          start_time + (imin / sr), tr.stats.channel)
    return amp


def get_response(filename, format=None, component=None):
    """
    Read response file and output PoleZeros object

    :param filename: name of the file to read
    :param format: 'GSE', 'JSON_PZ, 'SACPZ', 'STATIONXML' or None
    :param component: component to read, if STATIONXML
    """
    if format.upper() == 'GSE':
        paz = PAZ.read_gse_response(filename)
    elif format.upper() in ['JSON', 'JSON_PZ']:
        paz = PAZ.read_json_pz(filename)
    elif format.upper() == 'SACPZ':
        paz = PAZ.read_sac_pz(filename)
    elif format.upper() == 'STATIONXML':
        assert component is not None
        paz = PAZ.read_stationxml(filename, '*' + component)
    else:
        paz = PAZ.read_baillard_pz(filename)
    return paz


# def _read_JSON_PZ(filename):
#     """
#     read JSON paz format
#     """
#     with open(filename) as f:
#         paz = json.read(f)
#     A0 = 1
#     for p in paz.poles:
#         p = np.complex(p[0], p[1])
#         A0 *= (2*np.pi*paz['f_ref'] - p)
#     for z in paz.zeros:
#         z = np.complex(z[0], z[1])
#         A0 /= (2*np.pi*paz['f_ref'] - z)
#     A0 = _calc_A0(paz['poles'], paz['zeros'], paz['f_ref'],
#                   paz.get('A0', None))
#     paz['A0'] = A0
#     return paz
#
#
# def _calc_A0(poles, zeros, f_ref, proposed_A0=None):
#     """ Calculation normalization factor """
#     A0 = 1
#     for p in poles:
#         A0 *= (2*np.pi*f_ref - p)
#     for z in zeros:
#         A0 /= (2*np.pi*f_ref - z)
#     A0 = abs(A0)
#     if proposed_A0 is not None:
#         if A0/proposed_A0 < 0.99 or A0/proposed_A0 > 1.01:
#             log('file said A0 at {:.3g} Hz = {:.3g}, calculated {:.3g}'
#                 .format(f_ref, proposed_A0, A0), 'warning')
#     return A0
#
#
# def _read_Baillard_PZ(filename):
#     """
#     read Baillard PZ format
#     """
#     info = {}
#     with open(filename, 'r') as foc:
#         i = 0
#         a, info['poles'], info['zeros'] = [], [], []
#         for tline in foc:
#             if i < 4:
#                 a.append(float(tline))
#             elif i == 4:
#                 n_poles = float(tline)
#             elif i == 5:
#                 n_zeros = float(tline)
#             elif i >= 6 and i < (6 + n_poles):
#                 A = tline.split()
#                 info['poles'].append(np.complex(float(A[0]), float(A[1])))
#             elif i >= 6 + n_poles and i < 6 + n_poles + n_zeros:
#                 A = tline.split()
#                 info['zeros'].append(np.complex(float(A[0]), float(A[1])))
#             else:
#                 info['amplifier_gain'] = float(tline)
#             i += 1
#     info['sensor_gain'] = a[0]
#     info['f_ref'] = 1 / a[1]
#     info['sampling_rate'] = a[2]
#     info['A0'] = a[3]
#     _calc_A0(info['poles'], info['zeros'], info['f_ref'], info['A0'])
#     return info
#
#
# # read_GSE.m
# def _read_GSE(name=None):
#     """
#     read a GSE response file
#
#     :param name:    Name of GSE2 response file
#     :returns: Structure with all parameters
#     """
# # read_GSE.m:6
#     info = dict()
#     with open(name, 'r') as fic:
#         for line in fic:
#             if 'CAL' == line[:3]:
#                 info['station'] = line[6:10]
#                 info['type'] = line[20:29].trim()
#                 info['sensor_gain'] = float(line[29:40])
#                 info['sampling_rate'] = float(line[46:56])
#             else:
#                 if 'PAZ' == line[:3]:
#                     info['A0'] = float(line[11:25])
#                     num_poles = float(line[41:43])
#                     num_zeros = float(line[45])
#                     info['poles'] = []
#                     info['zeros'] = []
#                     for p in np.arange(num_poles):
#                         A = next(fic).split()
#                         info['poles'].extend(np.complex(float(A[0]),
#                                                         float(A[1])))
#                     for z in np.arange(num_zeros):
#                         A = next(fic).split()
#                         info['zeros'].extend(np.complex(float(A[0]),
#                                                         float(A[1])))
#                 else:
#                     if 'DIG' == line[:3]:
#                         info['amplifier_gain'] = float(line[9:23])
#                         info['digitizer'] = line[36:42].trim()
#                     else:
#                         continue
#     info['f_ref'] = 1
#     _calc_A0(info['poles'], info['zeros'], info['f_ref'], info['A0'])
#     return info



import setuptools

with open("README.rst", "r") as fh:
    long_description = fh.read()
    
version={}
with open("pspicker/version.py") as fp:
    exec(fp.read(),version)

setuptools.setup(
    name="pspicker",
    version=version['__version__'],
    author="Christian Baillard",
    author_email="crawford@ipgp.fr",
    description="Kurtosis-based P and S wave picker",
    long_description=long_description,
    long_description_content_type="text/x-rst; charset=UTF-8",
    url="https://github.com/WayneCrawford/pspicker",
    packages=setuptools.find_packages(),
    include_package_data=True,
    install_requires=[
          'obspy>=1.2',
          'numpy>=1.18',
          'scipy>=1.5',
          'verboselogs>=1.7',
          'matplotlib>=3.2',
          'pyyaml>=3.0',
          'jsonschema>=2.6',
          'jsonref>=0.2'
      ],
    entry_points={},
    python_requires='>=3.8',
    setup_requires=["pytest-runner"],
    tests_require=["pytest"],
    classifiers=(
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "Intended Audience :: Science/Research",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Physics"
    ),
    keywords='seismology'
)



"""
Routines to pick P and S waves using kurtosis extrema and to
"associate" them using clustering

Can also use the Kurtosis class independently
"""
from .pspicker import PSPicker
from .kurtosis import Kurtosis

__all__ = ['PSPicker', 'Kurtosis']



from .kurtosis_parameters import KurtosisParameters


class GlobalWindowParameters():
    """
    Global Window Parameters
    """
    def __init__(self,
                 kurtosis,
                 offsets,
                 distri_secs,
                 max_candidates=5,
                 end_cutoff=0.9):
        """
        Initialize Global Window Parameters

        :param kurtosis: KurtosisParameters dictionary
        :param distri_secs: Length in seconds of sliding window to use
            to find the densest pick time over all stations
        :param max_candidates: Number of pick candidates to use in evaluating
            the overall pick distribution
        :param offsets: Offsets in seconds from densest pick
            time for global (all station based) rewindowing [left, right]
        :param end_cutoff: What fraction of data (from start) to
            look at for global Kurtosis window. 1.0 looks everywhere
        """
        assert len(offsets) == 2, "len(offsets) != 2"

        self.kurtosis = KurtosisParameters(**kurtosis)
        assert len(self.kurtosis.frequency_bands) == 1
        self.offsets = offsets
        self.end_cutoff = end_cutoff
        self.distri_secs = distri_secs
        self.max_candidates = max_candidates

    def __str__(self):
        str = "GlobalWindowParameters:\n"
        str += f"    kurtosis = {self.kurtosis}\n"
        str += f"    distri_secs = {self.distri_secs}\n"
        str += f"    offsets = {self.offsets}\n"
        str += f"    end_cutoff = {self.end_cutoff}\n"
        str += f"    max_candidates = {self.max_candidates}\n"
        return str

    @classmethod
    def from_dict(cls, thedict):
        return cls(**thedict)


if __name__ == '__main__':
    pass



import copy

import numpy as np
from obspy.core.inventory.response import PolesZerosResponseStage

from .PAZ import _get_paz


def write_GSE2(fname, sta_code, ch, debug=False):
    """
    Write channel to the GSE2 format used by SEISAN

    GSE2 specification is found in GSE_2.0.pdf
    :param fname: name of file to write to
    :param sta_code: station code
    :param ch: obspy Channel objet
    """
    if debug:
        print(f'writing to {fname}\n')
        ch.response.plot(0.1, 'VEL')
    with open(fname, 'w') as f:
        f.write(_CAL2_line(ch, sta_code))
        f.write(_PAZ2_lines(ch))
        f.write(_DIG2_line(ch))


def _CAL2_line(ch, sta_code):
    """
    Returns Calibration Identification Line (CAL2)

    From Table 13, page 69
    Gives a rapid idea of sensitivity in passband
    """
    id = 'CAL2'.ljust(4)[:4]
    auxid = ''
    inst_type = ch.sensor.model
    # system gain (count/m) at ref freq (calfreq)
    calib, calfreq = _get_gain_disp(ch.response)
    print('CAL2 gain = {:.3g}'.format(abs(calib)))
    sens = 1.e9 / abs(calib)  # nm/c
    samprat = ch.sample_rate
    cal_line = '{} {} {} {} {} {:10.2e} {:7.3f} {:10.5f} {} {}\n'.format(
        id.ljust(4)[:4], sta_code.ljust(5)[:5], ch.code.ljust(3)[:3],
        auxid.ljust(4)[:4], inst_type.ljust(6)[:6],
        sens, 1 / calfreq, samprat,
        ch.start_date.strftime('%Y/%m/%d %H:%M'),
        ch.end_date.strftime('%Y/%m/%d %H:%M'))
    return cal_line


def _PAZ2_lines(ch):
    """
    Returns Pole-zero Response Section (PAZ2)

    From Table 14, page 70
    """
    paz = _get_paz(ch.response)
    # id = 'PAZ2'
    snum = paz.stage_sequence_number
    print(f'Last PAZ Stage number = {snum:d}')
    for st in ch.response.response_stages[0: snum]:
        print(st)
    # gain is in V/m
    gain, freq = _get_gain_disp(ch.response, 1, snum)
    # change poles/zeros to V/m also
    poles, zeros = _vel_to_disp(paz.poles, paz.zeros)
    nfact = _pz_norm_fact(poles, zeros, freq)
    gain_norm = gain * nfact
    print('PAZ gain = {abs(gain):.3g}')
    print('pz_norm_fact = {nfact:.3g}')
    norm_const = abs(gain_norm)/1e9
    deci = ''
    corr = ''
    npole = len(poles)
    nzero = len(zeros)
    descrip = paz.pz_transfer_function_type
    paz_header = f'PAZ2 {snum:2d} V {norm_const:15.8e} {deci.ljust(4)[:4]} '
    paz_header += f'{corr.ljust(8)[:8]} {npole:3d} {nzero:3d} '
    paz_header += f'{descrip.ljust(25)[:25]}\n'
    p_lines = ''
    for val in poles:
        p_lines += f' {val.real:15.8e} {val.imag:15.8e}\n'
    z_lines = ''
    for val in zeros:
        z_lines += f' {val.real:15.8e} {val.imag:15.8e}\n'
    return paz_header + p_lines + z_lines


def _DIG2_line(ch):
    """
    Returns Digitizer Response Section (DIG2)

    Assumes digitizer is right after the PZ stage(s)
    From Table 17, page 72
    """
    stages = [copy.deepcopy(stage) for stage in ch.response.response_stages
              if not isinstance(stage, PolesZerosResponseStage)]
    digi_stage = stages[0]
    id = 'DIG2'
    snum = digi_stage.stage_sequence_number
    sensitivity = digi_stage.stage_gain   # counts/input unit
    print(f'DIG2 gain = {sensitivity:.3g}')
    samprat = digi_stage.decimation_input_sample_rate
    descrip = digi_stage.description.replace('DIGITIZER', '').lstrip(' -')
    digi_line = f'{id.ljust(4)[:4]} {snum:2d} {sensitivity:15.8e} '
    digi_line += f'{samprat:11.5f} {descrip.ljust(25)[:25]}\n'
    # digi_line = id.ljust(4)[:4]  + ' ' + \
    #             '{:2d}'.format(snum)  + ' ' + \
    #             '{:15.8e}'.format(sensitivity)  + ' ' + \
    #             '{:11.5f}'.format(samprat)  + ' ' + \
    #             descrip.ljust(25)[:25]  + '\n'
    return digi_line


def _vel_to_disp(poles, zeros):
    """
    Convert velocity pole-zeros to displacement pole-zeros

    Removes a zero at zero frequency, if there is one, otherwise adds
    a poles at zero frequency
    """
    new_zeros = zeros.copy()
    new_poles = poles.copy()
    new_zeros.append(0. + 0.j)
    return new_poles, new_zeros


def _get_gain_disp(response, start_stage=None, end_stage=None):
    """
    Return stage(s) displacement gain
    """
    freq = response.instrument_sensitivity.frequency
    # resp is in outputs/inputs (full instrument = counts/m)
    gain = response.get_evalresp_response_for_frequencies(
        [freq], 'DISP', start_stage, end_stage)
    return gain[0], freq


def _pz_norm_fact(poles, zeros, ref_freq):
    """
    Returns pole-zero normalisation factor

    poles and zeros expressed in radians
    ref_freq in Hz
    """
    norm_fact = 1.
    for p in poles:
        norm_fact *= (2*np.pi*ref_freq*1j - p)
        print(abs(norm_fact))
    for z in zeros:
        norm_fact /= (2*np.pi*ref_freq*1j - z)
        print(abs(norm_fact))
    norm_fact = abs(norm_fact)
    return norm_fact



# Generated with SMOP  0.41-beta
# from smop.libsmop import *
# readmain.m


class ChannelMappingRules:
    """
    Rules for mapping channels
    """
    def __init__(self, compZ='Z3', compE='E2X', compN='N1Y', compH='HF',
                 S_write_cmp='N', P_write_cmp='Z',
                 P_write_phase='Pg', S_write_phase='Sg',
                 band_order='GFDCEHSBMLV'):
        """
        :param compZ: comp characters to assign to Z component
        :param compE: comp characters to assign to E component
        :param compN: comp characters to assign to N component
        :param compH: comp characters to assign to H component
        :param S_write_cmp: save S picks to this component.  Should be
            one of 'ZNEH': output will be written the the component matching
            this value using the comp? regexp above
        :param P_write_cmp: save P picks to this component.  Should be
            one of 'ZNEH'
        :param S_write_phase: call S picks this phase name
        :param P_write_phase: call P picks this phase name
        :param band_order: order of preference for band codes (if component
            choice does not yield unique selection)
        """
        assert S_write_cmp in 'ZNEH',\
            f"S_write_cmp = {S_write_cmp}, not in 'ZNEH'"
        assert P_write_cmp in 'ZNEH',\
            f"P_write_cmp = {P_write_cmp}, not in 'ZNEH'"
        self.compZ = compZ
        self.compE = compE
        self.compN = compN
        self.compH = compH
        self.S_write_cmp = S_write_cmp
        self.P_write_cmp = P_write_cmp
        self.P_write_phase = P_write_phase
        self.S_write_phase = S_write_phase
        self.band_order = band_order

    def __str__(self):
        str = "ChannelMappingRules\n"
        str += f"    compZ = {self.compZ}\n"
        str += f"    compE = {self.compE}\n"
        str += f"    compN = {self.compN}\n"
        str += f"    compH = {self.compH}\n"
        str += f"    S_write_cmp = {self.S_write_cmp}\n"
        str += f"    P_write_cmp = {self.P_write_cmp}\n"
        str += f"    P_write_phase = {self.P_write_phase}\n"
        str += f"    S_write_phase = {self.S_write_phase}\n"
        return str

    @classmethod
    def from_dict(cls, thedict):
        return(cls(**thedict))



class PickerRunParameters():
    """
    Parameters associated with the run of one event
    """
    def __init__(self,
                 database_filename,
                 wavefile=None,
                 stream=None,
                 channel_maps=None,
                 first_time=None,
                 last_time=None,
                 t_begin=None):
        """
        :param database_filename: database file to read
        :param wavefile: name of the file containing the waveforms
        :param stream: all of the data waveforms
        :param channel_maps: a dict of ChannelMapping objects with station
            names as key
        :param t_begin: time of reference for overall_distri
        :param first_time: never look before this time
        :param last_time: never look after this time
        """
        self.database_filename = database_filename
        self.channel_maps = channel_maps
        self.wavefile = wavefile
        self.stream = stream
        self.first_time = first_time
        self.last_time = last_time
        self.t_begin = t_begin

    @property
    def stations(self):
        return [s for s in self.channel_maps.keys()]

    @property
    def n_stations(self):
        return len(self.stations)


if __name__ == '__main__':
    pass
